/*
 *	A Test of Lestrade: Indexing of Songs in the Virtual Songbook
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 */

#include "ucw/lib.h"
#include "ucw/getopt.h"
#include "ucw/conf.h"
#include "ucw/clists.h"
#include "ucw/fastbuf.h"
#include "charset/charconv.h"
#include "charset/fb-charconv.h"
#include "lestrade/lestrade.h"
#include "lestrade/tagged-text.h"
#include "custom/custom.h"

#include <stdio.h>
#include <stdlib.h>
#include <ftw.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

static int verbose;
static uns max_results = 5;
static clist dir_list, idx_list;

static void NONRET
usage(void)
{
  fputs("\
Usage:\n\
\n\
To build an index:\n\
  songsearch [-v] -b <index> <directory>\n\
\n\
To search:\n\
  songsearch [-v] [-d <dir>]* [-i <index>]* [-n <max-matches>] <query>\n\
", stderr);
  exit(1);
}

struct name {
  cnode n;
  byte *name;
};

static clist file_list;

static void
add_name(clist *l, byte *n)
{
  struct name *x = cf_malloc(sizeof(*x));
  x->name = cf_strdup(n);
  clist_add_tail(l, &x->n);
}

static struct fastbuf *
song_read(byte *name)
{
  if (verbose > 1)
    msg(L_INFO, "Reading %s", name);
  struct fastbuf *infile = bopen(name, O_RDONLY, 65536);
  struct fastbuf *in = fb_wrap_charconv_in(infile, CONV_CHARSET_ISO_8859_2, CONV_CHARSET_UTF8);
  struct fastbuf *out = fbmem_create(1024);
  byte line[1024];

  for (;;)
    {
      if (!bgets(in, line, sizeof(line)))
	die("Truncated header in %s", name);
      if (!line[0])
	break;
      if (!strncasecmp(line, "Name:", 5))
	{
	  bputc(out, 0x90 + WT_TITLE);
	  bputs(out, line+5);
	}
      else if (!strncasecmp(line, "Author:", 7))
	{
	  bputc(out, 0x90 + WT_AUTHOR);
	  bputs(out, line+7);
	}
    }

  bputc(out, 0x90 + WT_TEXT);
  while (bgets(in, line, sizeof(line)))
    if (line[0] && line[0] != '!')
      bputsn(out, line);

  bclose(in);
  bclose(infile);
  bflush(out);
  struct fastbuf *o = fbmem_clone_read(out);
  bclose(out);
  return o;
}

static struct fastbuf *
song_enum(struct card_attrs *attrs, void *priv)
{
  struct name *n = clist_head(priv);
  if (!n)
    return NULL;
  clist_remove(&n->n);
  struct fastbuf *fb = song_read(n->name);
  strcpy(attrs->name, n->name);
  return fb;
}

static struct fastbuf *
song_fetch(struct card_attrs *attrs, void *priv UNUSED)
{
  return song_read(attrs->name);
}

static int
ftw_callback(const char *file, const struct stat *sb UNUSED, int flag)
{
  if (flag != FTW_F)
    return 0;
  const char *sep = strrchr(file, '/');
  sep = (sep ? sep+1 : file);
  if (sep[0] == '.')
    return 0;
  if (strlen(file) >= CARD_NAME_LENGTH)
    die("Name `%s' is too long for us to grasp", file);
  add_name(&file_list, (byte*)file);
  return 0;
}

static clist *
scan_files(byte *dir)
{
  clist_init(&file_list);
  if (ftw(dir, ftw_callback, 64))
    die("Error scanning %s: %m", dir);
  return &file_list;
}

static void
mkindex(byte *idxname, byte *srcdir)
{
  struct custom_index_hdr cih;
  strcpy(cih.directory, srcdir);
  if (verbose)
    cf_set("Indexer.Verbose=1");
  lestrade_make_index(idxname, song_enum, scan_files(srcdir), &cih, sizeof(cih));
}

static uns spc;

static const byte *
print_range(struct fastbuf *out, const byte *curr, const byte *until)
{
  while (curr < until)
    {
      uns u;
      curr = get_tagged_char(curr, &u);
      if (u < 0x80000000)
	{
	  if (u <= ' ')
	    spc = 1;
	  else
	    {
	      if (spc)
		{
		  bputc(out, ' ');
		  spc = 0;
		}
	      bput_utf8(out, u);
	    }
	}
    }
  if (spc)
    {
      bputc(out, ' ');
      spc = 0;
    }
  return curr;
}

static void
print_interval(struct fastbuf *out, struct lestrade_matches *m, struct lestrade_word_pos *from, struct lestrade_word_pos *to)
{
  const byte *curr = m->text + from->pos;
  spc = 0;
  while (from < to)
    {
      curr = print_range(out, curr, m->text + from->pos);
      if (from->word_index != 0xff)
	bputc(out, '<');
      curr = print_range(out, curr, m->text + from->pos + from->length);
      if (from->word_index != 0xff)
	bputc(out, '>');
      from++;
    }
}

static void
print_title(struct fastbuf *out, struct lestrade_matches *m, uns wt)
{
  for (struct lestrade_word_pos *wp = m->words; wp->word_type != 0xff; wp++)
    if (wp->word_type == wt)
      {
	struct lestrade_word_pos *start = wp;
	while (wp->word_type == wt)
	  wp++;
	print_interval(out, m, start, wp);
	return;
      }
  bputs(out, "???");
}

static void
print_matches(struct fastbuf *out, struct lestrade_matches *m)
{
  bputc(out, '\t');
  print_title(out, m, WT_AUTHOR);
  bputs(out, ": ");
  print_title(out, m, WT_TITLE);
  bputc(out, '\n');
  struct lestrade_match *best = NULL;
  for (uns i=0; i<m->num_matches; i++)
    if (m->words[m->matches[i].pos].word_type == WT_TEXT &&
	(!best || m->matches[i].weight > best->weight))
      best = &m->matches[i];
  if (!best)
    return;
  int l = best->pos;
  while (l > (int)best->pos-8 && l > 0 && m->words[l-1].word_type == WT_TEXT)
    l--;
  int r = best->pos;
  while (r < (int)best->pos+8 && m->words[r+1].word_type == WT_TEXT)
    r++;
  bputc(out, '\t');
  print_interval(out, m, m->words+l, m->words+r+1);
  bputc(out, '\n');
}

static void
query(char **argv, int argc)
{
  struct lestrade_query *Q = lestrade_new_query();
  Q->fetcher = song_fetch;
  if (verbose)
    Q->trace_mask = (verbose > 1) ? ~0U : TRACE_GENERAL;
  Q->max_results = max_results;

  Q->expr = lestrade_new_match(Q);
  for (int i=0; i<argc; i++)
    {
      byte *c = argv[i];
      int sense = 0;
      if (*c == '+')
	{
	  sense = 1;
	  c++;
	}
      else if (*c == '-')
	{
	  sense = -1;
	  c++;
	}
      struct fastbuf fbb;
      fbbuf_init_read(&fbb, c, strlen(c), 0);
      struct fastbuf *fbc = fb_wrap_charconv_in(&fbb, CONV_CHARSET_ISO_8859_2, CONV_CHARSET_UTF8);
      byte at[1024];
      bgets(fbc, at, sizeof(at));
      bclose(fbc);
      struct lestrade_atom *a = lestrade_add_atom(Q, Q->expr, at);
      a->sense = sense;
      a->accent_mode = ACCENT_MODE_STRIP;
      a->morph_mode = MORPH_MODE_PREACCENT;
    }

  if (lestrade_init_query(Q) < 0)
    die("Error: %s", Q->error_msg);

  struct name *n;
  CLIST_WALK(n, idx_list)
    {
      struct lestrade_index *idx;
      int err;
      if (err = lestrade_open_index(n->name, &idx))
	die("Cannot open index %s: error %d", n->name, err);
      lestrade_eval_index(Q, idx);
      lestrade_close_index(idx);
    }
  if (!clist_empty(&dir_list))
    {
      if (lestrade_eval_cards_init(Q) < 0)
	die("Error: %s", Q->error_msg);
      CLIST_WALK(n, dir_list)
	lestrade_eval_cards(Q, song_enum, scan_files(n->name));
      lestrade_eval_cards_cleanup(Q);
    }

  lestrade_get_results(Q);
  if (!Q->num_results)
    {
      puts("Not found.");
      return;
    }
  struct fastbuf *stdout = bfdopen_shared(1, 4096);
  struct fastbuf *out = fb_wrap_charconv_out(stdout, CONV_CHARSET_UTF8, CONV_CHARSET_ISO_8859_2);
  if (lestrade_eval_detailed_init(Q) < 0)
    die("Error: %s", Q->error_msg);

  for (uns i=0; i<Q->num_results; i++)
    {
      bprintf(out, "%d. [%5d] %s\n", i+1, Q->results[i].quality, Q->results[i].attrs.name);
      struct lestrade_matches *m = lestrade_eval_detailed(Q, &Q->results[i].attrs);
      print_matches(out, m);
    }

  lestrade_eval_detailed_cleanup(Q);
  bclose(out);
  bclose(stdout);
  lestrade_free_query(Q);
}

int main(int argc, char **argv)
{
  byte *build_index = NULL;

  log_init(argv[0]);
  clist_init(&dir_list);
  clist_init(&idx_list);
  int opt;
  while ((opt = cf_getopt(argc, argv, CF_SHORT_OPTS "b:d:i:n:v", CF_NO_LONG_OPTS, NULL)) >= 0)
    switch (opt)
      {
      case 'd':
	add_name(&dir_list, optarg);
	break;
      case 'i':
	add_name(&idx_list, optarg);
	break;
      case 'b':
	if (build_index)
	  usage();
	build_index = optarg;
	break;
      case 'n':
	max_results = atol(optarg);
	break;
      case 'v':
	verbose++;
	break;
      default:
	usage();
      }

  lestrade_init(NULL);
  if (optind < argc && !build_index && (!clist_empty(&dir_list) || !clist_empty(&idx_list)))
    query(argv+optind, argc-optind);
  else if (optind == argc-1 && build_index && clist_empty(&dir_list) && clist_empty(&idx_list))
    mkindex(build_index, argv[optind]);
  else
    usage();

  lestrade_cleanup();
  return 0;
}
