/*
 *	Sherlock Indexer -- Character Class Table for Lexical Mapping
 *
 *	(c) 2003--2005 Martin Mares <mj@ucw.cz>
 *	(c) 2006 Robert Spalek <robert@ucw.cz>
 *	(c) 2009 Pavel Charvat <pchar@ucw.cz>
 */

#include "ucw/lib.h"
#include "lestrade/lestrade.h"
#include "ucw/conf.h"
#include "ucw/slists.h"
#include "ucw/chartype.h"
#include "ucw/unicode.h"
#include "charset/unicat.h"
#include "lestrade/alphabet.h"

#include <string.h>

byte alpha_class[65536];

struct unirange {
  u16 min, max;
};

struct alpha_entry {
  snode n;
  struct unirange range;
  enum alphabet_class class;
};

static slist alpha_conf;
static uns alpha_built;

static uns
uni_parser(byte *c, u16 *up)
{
  uns u;
  byte *cc = (byte*)utf8_get(c, &u);
  if (*cc || u == UNI_REPLACEMENT)
    {
      for (uns i=0; i<4; i++)
	if (!Cxdigit(c[i]))
	  return 1;
	else
	  u = (u << 4) | Cxvalue(c[i]);
      if (c[4])
	return 1;
    }
  *up = u;
  return 0;
}

static byte *
unirange_parser(byte *s, struct unirange *ur)
{
  byte *c;
  if ((c = strchr(s, '-')) && c > s)
    {
      *c++ = 0;
      if (uni_parser(s, &ur->min) || uni_parser(c, &ur->max))
	goto err;
    }
  else
    {
      if (uni_parser(s, &ur->min))
	goto err;
      ur->max = ur->min;
    }
  if (ur->min > ur->max)
    return "Invalid code range (min>max)";
  return NULL;

 err:
  return "Incorrect syntax of a code range";
}

static void
alpha_build(void)
{
  bzero(alpha_class, sizeof(alpha_class));
  SLIST_FOR_EACH(struct alpha_entry *, e, alpha_conf)
    for (uns i=e->range.min; i<=e->range.max; i++)
      {
	uns c = e->class;
	if (c == AC_INHERIT)
	  {
	    uns u = Ucategory(i);
	    if (u & _U_LETTER)
	      c = AC_ALPHA;
	    else if (u & _U_DIGIT)
	      c = AC_DIGIT;
	    else if (u & (_U_SPACE | _U_CTRL))
	      c = AC_SPACE;
	    else if (u & _U_LIGATURE)
	      c = AC_LIGATURE;
	    else
	      c = AC_PUNCT;
	  }
	alpha_class[i] = c;
      }
  alpha_built = 1;
}

static byte *
parse_class(uns n, byte **pars, void *pclass)
{
  alpha_built = 0;
  for (uns i=0; i<n; i++)
    {
      struct alpha_entry *e = cf_malloc_zero(sizeof(*e));
      byte *err;
      if (err = unirange_parser(pars[i], &e->range))
	return err;
      e->class = (enum alphabet_class) pclass;
      /* Rather crude journaling of slists */
      CF_JOURNAL_VAR(alpha_conf.last);
      if (alpha_conf.last)
	CF_JOURNAL_VAR(alpha_conf.last->next);
      else
	CF_JOURNAL_VAR(alpha_conf.head);
      slist_add_tail(&alpha_conf, &e->n);
    }
  return NULL;
}

static struct cf_section alphabet_config = {
  CF_ITEMS {
    CF_PARSER("Inherit", (void*) AC_INHERIT, parse_class, CF_ANY_NUM),
    CF_PARSER("Space", (void*) AC_SPACE, parse_class, CF_ANY_NUM),
    CF_PARSER("Alpha", (void*) AC_ALPHA, parse_class, CF_ANY_NUM),
    CF_PARSER("Punct", (void*) AC_PUNCT, parse_class, CF_ANY_NUM),
    CF_PARSER("Break", (void*) AC_BREAK, parse_class, CF_ANY_NUM),
    CF_PARSER("Singleton", (void*) AC_SINGLETON, parse_class, CF_ANY_NUM),
    CF_END
  }
};

static void CONSTRUCTOR
alphabet_preinit(void)
{
  cf_declare_section("Alphabet", &alphabet_config, 0);
}

void
alphabet_init(void)
{
  if (!alpha_built)
    {
      alpha_build();
      if (alpha_class['A'] == AC_SPACE)
        die("Letter `A' is configured as a space. You probably forgot to load the configuration file.");
    }
}
